from argparse import SUPPRESS
from argparse import ArgumentParser
from argparse import Namespace
from argparse import RawDescriptionHelpFormatter
from argparse import _SubParsersAction as SubParsers
from collections.abc import Callable
from os import environ
from pathlib import Path
from subprocess import run
from sys import exit
from sys import stderr
from sys import stdin
from sys import version_info
from tempfile import NamedTemporaryFile
from textwrap import dedent
from typing import Any

from .config import __metadata__
from .ezchlog import EzChLog


class ClapArgumentParser(ArgumentParser):
    def __init__(self, *args, metadata: dict | None = None, **kwargs):
        kwargs['formatter_class'] = RawDescriptionHelpFormatter
        if metadata:
            kwargs['description'] = dedent(
                f"""
                {metadata['summary']} ({metadata['name']} {metadata['version']})
                {metadata['author_email']}
                """,
            )
            kwargs['prog'] = metadata['name']
            kwargs['add_help'] = False
        super().__init__(*args, **kwargs)
        if metadata:
            self.add_argument('-h', '--help', action='help', default=SUPPRESS, help="Print help")
            self.add_argument('-V', '--version', action='version', version=f"{metadata['name']} {metadata['version']}", help="Print version")

    def format_help(self):
        formatter = self._get_formatter()
        formatter.add_text(self.description)
        formatter.add_usage(self.usage, self._actions, self._mutually_exclusive_groups)
        order = {  # builtin titles
            'positional arguments': '1',
            'options': '2',
        }
        for action_group in sorted(self._action_groups, key=lambda ag: order.get(ag.title, '0') + ag.title):
            formatter.start_section(action_group.title)
            formatter.add_text(action_group.description)
            formatter.add_arguments(action_group._group_actions)
            formatter.end_section()
        formatter.add_text(self.epilog)
        return formatter.format_help()


class Parser:
    def __init__(self, ezchlog: EzChLog) -> None:
        self.log = ezchlog
        self.parser = ClapArgumentParser(metadata=__metadata__)
        subparsers = self.parser.add_subparsers(required=True, dest='command', metavar='COMMAND', title='commands')
        parser_show_config = self._add_command(subparsers, 'showconfig', self.show_config_command, with_dry_run=False)
        parser_show_config.add_argument(
            dest='config_key',
            nargs='?',
            metavar='key',
            default='',
            help="Only show this config key value.",
        )
        parser_add = self._add_command(subparsers, 'add', self.add_command)
        parser_add.add_argument(
            '--no-git',
            action='store_true',
            help="Disable dealing with git index",
        )
        parser_add.add_argument(
            '-b',
            '--branch',
            dest='create_branch',
            action='store_true',
            help="Create the git branch along the changelog",
        )
        parser_add.add_argument(
            '-p',
            '--path',
            dest='use_paths',
            action='store_true',
            help="Output paths",
        )
        parser_add.add_argument(
            'message',
            help="Your message. Use '-' to open an editor instead.",
        )
        Category = self.log.cfg.category_class
        CategoryDef = Category[self.log.cfg.category_default]
        category_names = [cat.name for cat in Category]
        parser_add.add_argument(
            'cat',
            nargs='?',
            metavar='category',
            default=CategoryDef,
            type=lambda s: getattr(Category, s) if s else CategoryDef,
            choices=list(Category),
            help=f"Choose one of {', '.join(category_names)}. Defaut to {CategoryDef.name}.",
        )
        parser_add.add_argument(
            'ref',
            nargs='?',
            metavar='reference',
            default='',
            help="Reference for the log. Default is empty.",
        )
        parser_commit = self._add_command(subparsers, 'commit', self.commit_command)
        parser_commit.add_argument(
            'partlog_path',
            nargs='?',
            metavar='partlog',
            default=None,
            help="Main part log path if this is ambiguous",
        )
        parser_list = self._add_command(subparsers, 'list', self.list_command, with_dry_run=False)
        parser_list.add_argument(
            '-p',
            '--path',
            dest='use_paths',
            action='store_true',
            help="Output paths",
        )
        parser_merge = self._add_command(subparsers, 'merge', self.merge_command)
        parser_merge.add_argument(
            '--no-git',
            action='store_true',
            help="Disable dealing with git index",
        )
        parser_merge.add_argument(
            '-d',
            '--with-date',
            action='store_true',
            help="Add the current ISO date to the version title",
        )
        parser_merge.add_argument(
            '-l',
            '--link',
            dest='link',
            default='',
            help="Link to commits diff associated with the version",
        )
        parser_merge.add_argument(
            '-s',
            '--subtitle',
            dest='subtitle',
            default='',
            help="Add some markdown text under the version title",
        )
        parser_merge.add_argument(
            'version',
            metavar='next_ver',
            help="The next version",
        )

        def help_func(opts):
            """Print this message or the help of the given subcommand(s)"""
            self.parser.print_help()

        self._add_command(subparsers, 'help', help_func)

    def _add_command(self, subparsers: SubParsers, name: str, func: Callable, *, with_dry_run: bool = True) -> ArgumentParser:
        help_text = getattr(func, '__doc__', None)
        sub_parser = subparsers.add_parser(name, help=help_text, description=help_text)
        if with_dry_run:
            sub_parser.add_argument(
                '-n',
                '--dry-run',
                action='store_true',
                help="Dry-run the command",
            )
        sub_parser.set_defaults(func=func)
        return sub_parser

    def show_config_command(self, opts: Namespace) -> None:
        """Show configuration"""

        def format_value(value: Any | None, *, in_toml: bool = True) -> str:
            match value:
                case Path():
                    return format_value(
                        str(value.absolute().relative_to(self.log.cfg.root_dir.absolute())),
                        in_toml=in_toml,
                    )
                case str():
                    if in_toml:
                        return f'"""\n{value}\n"""' if '\n' in value else f'"{value}"'
                    else:
                        return value
                case list():  # assume items are strings
                    if in_toml:
                        return '[' + ', '.join(f'"{item}"' for item in value) + ']'
                    else:
                        return '\n'.join(str(item) for item in value)
                case bool():
                    return str(value).lower()
                case _:
                    return f'{value}'

        if opts.config_key:
            value = dict(self.log.cfg).get(opts.config_key)
            print(format_value(value, in_toml=False))  # noqa: T201
        else:
            for key, value in self.log.cfg:
                print(f'{key} = {format_value(value)}')  # noqa: T201

    def open_editor(self, file_ext: str, default_message: str) -> str:
        editor = environ.get('EZCHLOG_EDITOR', environ.get('EDITOR', environ.get('VISUAL', 'vim')))
        if not editor or not stdin.isatty():
            raise Exception(f"Cannot run editor '{editor}'")
        with NamedTemporaryFile(mode='w+', encoding='utf-8', suffix=f'.{file_ext}', delete=False) as f:
            f.write(default_message)
            f.flush()
            try:
                run([editor, f.name])
                f.seek(0)
                return '\n'.join(line for line in f.read().split('\n') if not line.startswith('#'))
            finally:
                Path(f.name).unlink()

    @staticmethod
    def _relative_to_walk_up(src_path: Path, dest_path: Path) -> Path:
        src_full_path = src_path.absolute()
        if list(version_info)[:2] >= [3, 12]:
            try:
                kwargs = dict(walk_up=True)  # split in two to please parsers/linters
                return src_full_path.relative_to(dest_path.absolute(), **kwargs)
            except ValueError:
                return src_full_path
        else:  # simulate walk_up for python < 3.12
            pp = Path()
            cur_path = dest_path.absolute()
            while not cur_path.is_mount():
                try:
                    return pp / src_full_path.relative_to(cur_path)
                except ValueError:
                    pp /= '..'
                    cur_path = cur_path.parent
            return src_full_path

    def add_command(self, opts: Namespace) -> None:
        """Add a changelog part file"""
        if opts.message == '-':
            opts.message = self.open_editor(
                'md',
                """
# This a markdown log file.
# Any comment will be removed.
# An empty file will abort.
""",
            )
        message = opts.message.strip()
        if not message:
            raise Exception("Aborted")
        no_git = self.log.cfg.no_git or opts.no_git
        create_branch = not no_git and opts.create_branch
        add_to_index = not no_git and bool(self.log.cfg.git_dir)
        p, md_message = self.log.add(dry_run=opts.dry_run, message=message, cat=opts.cat, ref=opts.ref, create_branch=create_branch, add_to_index=add_to_index)
        print(self._relative_to_walk_up(self.log.cfg.log_dir / p, Path.cwd()) if opts.use_paths else f'{p.parent!s: <20}{p.stem}')  # noqa: T201
        if opts.dry_run:
            print(md_message)  # noqa: T201

    def commit_command(self, opts: Namespace) -> None:
        """Commit using part changelog files content"""
        message = self.log.commit(dry_run=opts.dry_run, partlog_path=opts.partlog_path)
        if opts.dry_run:
            print(message, end='')  # noqa: T201

    def list_command(self, opts: Namespace) -> None:
        """List changelog files"""
        for p in self.log.list():
            print(self._relative_to_walk_up(self.log.cfg.log_dir / p, Path.cwd()) if opts.use_paths else f'{p.parent!s: <20}{p.stem}')  # noqa: T201

    def merge_command(self, opts: Namespace) -> None:
        """Merge changelog files into main changelog file under a version"""
        with_date = self.log.cfg.with_date or opts.with_date
        no_git = self.log.cfg.no_git or opts.no_git
        update_index = not no_git and bool(self.log.cfg.git_dir)
        changelog = self.log.merge(
            dry_run=opts.dry_run,
            next_version=opts.version,
            link=opts.link,
            subtitle=opts.subtitle,
            with_date=with_date,
            update_index=update_index,
        )
        if opts.dry_run:
            print(changelog)  # noqa: T201

    def parse(self, args: list[str] | None = None) -> None:
        try:
            opts = self.parser.parse_args(args=args)
            opts.func(opts)
        except Exception as e:
            if environ.get('DEBUG'):
                from traceback import print_exc

                print_exc()
            else:
                print(f"Error: {e}", file=stderr)  # noqa: T201
            exit(1)


def run_cli() -> None:
    Parser(EzChLog()).parse()



from collections.abc import Iterator
from functools import cached_property
from importlib.metadata import PackageNotFoundError
from importlib.metadata import metadata
from os import environ
from pathlib import Path
from typing import Any
from typing import TypedDict

try:
    from tomllib import load  # type: ignore[import-not-found]
except ModuleNotFoundError:
    from tomli import load  # type: ignore
try:
    _metadata_message = metadata(__package__ or __name__)
    __metadata__ = _metadata_message.json
except PackageNotFoundError:
    __metadata__ = dict(
        name=__name__,
        version='unknown',
        summary=__name__,
        author_email='unknown',
    )


class ConfigParams(TypedDict):
    log_file: str
    log_dir: str
    category_list: list[str]
    category_default: str
    default_changelog: str
    with_date: bool
    no_git: bool
    branch_format: str
    branch_separator: str
    branch_lowercase_for: list[str]


DEFAULTS = ConfigParams(
    log_file='CHANGELOG.md',
    log_dir='_CHANGELOGS',
    category_list=[
        'Breaking:💣',
        'Security:🛡️',
        'Fixed:🔧',
        'Changed:📝',
        'Added:➕',
        'Removed:➖',
        'Deprecated:⚠️',
    ],
    category_default='Changed',
    default_changelog="""\
# Changelog
All notable changes to this project will be documented in this file.

The format is based on [Keep a Changelog](http://keepachangelog.com/en/1.1.0/)
and this project adheres to [Semantic Versioning](http://semver.org/spec/v2.0.0.html).
""",
    with_date=False,
    no_git=False,
    branch_format="%{ref}%{sep}%{name}",  # allowed variables: ref, cat, name, sep
    branch_separator="_",
    branch_lowercase_for=['cat', 'name'],  # allowed variables: ref, cat, name
)


class CategoryMeta(type):
    _lst = dict[str, 'Category']()

    def __iter__(cls) -> Iterator['Category']:
        return iter(cls._lst.values())

    def __getitem__(cls, item: str) -> 'Category':
        return cls._lst[item]

    def set_list(cls, category_list: list[str]) -> None:
        lst = dict[str, Category]()
        for name, icon in ((values[0], ''.join(values[1:])) for cat in category_list if (values := cat.split(':'))):
            lst[name] = cls(name, icon)
        cls._lst = lst


class Category(metaclass=CategoryMeta):
    def __init__(self, name: str, icon: str) -> None:
        self.name = name
        self.icon = icon

    def to_definition(self) -> str:
        values = [self.name]
        if self.icon:
            values.append(self.icon)
        return ':'.join(values)


class Config:
    curr_dir: Path
    log_file: Path
    log_dir: Path
    category_list: list[str]
    category_default: str
    default_changelog: str
    with_date: bool
    no_git: bool
    branch_format: str
    branch_separator: str
    branch_lowercase_for: list[str]

    def __init__(self) -> None:
        self.curr_dir = Path.cwd().resolve()
        cfg: dict[str, Any] = {}
        if self.pyproject:
            with self.pyproject.open('rb') as f:
                cfg |= load(f).get('tool', {}).get('ezchlog', {})
        if self.ezchlogconf:
            with self.ezchlogconf.open('rb') as f:
                cfg |= load(f)
        self.log_file = Path(environ.get('EZCHLOG_LOG_FILE', cfg.get('log_file', DEFAULTS['log_file'])))
        if not self.log_file.is_absolute():
            self.log_file = self.root_dir / self.log_file
        self.log_dir = Path(environ.get('EZCHLOG_LOG_DIR', cfg.get('log_dir', DEFAULTS['log_dir'])))
        if not self.log_dir.is_absolute():
            self.log_dir = self.root_dir / self.log_dir
        raw_categories: str | list[str] = environ.get('EZCHLOG_CATEGORY_LIST', cfg.get('category_list', DEFAULTS['category_list']))
        self.category_list = raw_categories.split(',') if isinstance(raw_categories, str) else raw_categories
        self.category_default = environ.get('EZCHLOG_CATEGORY_DEFAULT', cfg.get('category_default', DEFAULTS['category_default']))
        self.default_changelog = environ.get('EZCHLOG_DEFAULT_CHANGELOG', cfg.get('default_changelog', DEFAULTS['default_changelog']))
        self.with_date = environ.get('EZCHLOG_WITH_DATE', cfg.get('with_date', DEFAULTS['with_date'])) in (True, 'true', 'True', '1', 'on')
        self.no_git = environ.get('EZCHLOG_NO_GIT', cfg.get('no_git', DEFAULTS['no_git'])) in (True, 'true', 'True', '1', 'on')
        self.branch_format = environ.get('EZCHLOG_BRANCH_FORMAT', cfg.get('branch_format', DEFAULTS['branch_format']))
        self.branch_separator = environ.get('EZCHLOG_BRANCH_SEPARATOR', cfg.get('branch_separator', DEFAULTS['branch_separator']))
        raw_br_lc_for: str | list[str] = environ.get('EZCHLOG_BRANCH_LOWERCASE_FOR', cfg.get('branch_lowercase_for', DEFAULTS['branch_lowercase_for']))
        self.branch_lowercase_for = raw_br_lc_for.split(',') if isinstance(raw_br_lc_for, str) else raw_br_lc_for

    @cached_property
    def ezchlogconf(self) -> Path | None:
        p_root = self.curr_dir
        for p in [p_root] + list(p_root.parents):
            pf = p / '.ezchlog.toml'
            if pf.is_file():
                return pf
        else:
            return None

    @cached_property
    def pyproject(self) -> Path | None:
        p_root = self.curr_dir
        for p in [p_root] + list(p_root.parents):
            pf = p / 'pyproject.toml'
            if pf.is_file():
                return pf
        else:
            return None

    @cached_property
    def editorconfig(self) -> Path | None:
        p_root = self.curr_dir
        for p in [p_root] + list(p_root.parents):
            pf = p / '.editorconfig'
            if pf.is_file():
                return pf
        else:
            return None

    @cached_property
    def git_dir(self) -> Path | None:
        p_root = self.curr_dir
        for p in [p_root] + list(p_root.parents):
            pd = p / '.git'
            if pd.is_dir():
                return pd
        else:
            return None

    @cached_property
    def root_dir(self) -> Path:
        for attr in ('ezchlogconf', 'pyproject', 'editorconfig', 'git_dir'):
            if path := getattr(self, attr):
                return path.parent
        else:
            return self.curr_dir

    @cached_property
    def category_class(self) -> type[Category]:
        Category.set_list(self.category_list)
        return Category

    def __iter__(self) -> Iterator[tuple[str, Path | str | list[str] | bool]]:
        category_list = list(cat.to_definition() for cat in self.category_class)
        d = dict[str, Path | str | list[str] | bool](
            log_file=self.log_file,
            log_dir=self.log_dir,
            category_list=category_list,
            category_default=self.category_default,
            default_changelog=self.default_changelog,
            with_date=self.with_date,
            no_git=self.no_git,
            branch_format=self.branch_format,
            branch_separator=self.branch_separator,
            branch_lowercase_for=self.branch_lowercase_for,
        )
        return iter(d.items())



from datetime import datetime
from datetime import timezone
from pathlib import Path
from re import match
from re import sub
from subprocess import run
from unicodedata import normalize

from .config import Category
from .config import Config


class StrChainSub:
    def __init__(self, s: str) -> None:
        self.s = s

    def sub(self, regex: str, replacement: str) -> 'StrChainSub':
        return StrChainSub(sub(regex, replacement, self.s))

    def __str__(self) -> str:
        return self.s


class EzChLog:
    def __init__(self) -> None:
        self.cfg = Config()

    @classmethod
    def get_sane_str(cls, s: str, sep: str = '_', max_size: int = 50) -> str:
        s = normalize('NFD', s).encode('ascii', 'ignore').decode('utf-8')
        stripped = ''.join(set((sep, '-')))
        return (
            str(
                StrChainSub(s).sub(r'\s+', sep).sub(r'[^-a-zA-Z0-9_]', '').sub(f'{sep}+', sep),
            )[:max_size]
            .strip()
            .strip(stripped)
        )

    @classmethod
    def get_slug(cls, s: str) -> str:
        return cls.get_sane_str(s.lower())

    def get_branch_name(self, cat: Category, ref: str, name: str) -> str:
        sep = self.cfg.branch_separator
        lowercase_for = self.cfg.branch_lowercase_for
        br_name = self.cfg.branch_format
        for key, (value, size) in {'ref': (ref, 10), 'cat': (cat, 20), 'name': (name, 30)}.items():
            value_str = getattr(value, 'name', str(value))
            value = value_str.lower() if key in lowercase_for else value_str
            value = self.get_sane_str(value, sep, size)
            br_name = br_name.replace('%{' + key + '}', value)
        br_name = br_name.replace('%{sep}', sep)
        return self.get_sane_str(br_name, sep)

    def run_command(self, *args: str, **kwargs) -> str:
        ret = run(args, cwd=str(self.cfg.root_dir), capture_output=True, encoding='utf-8', **kwargs)
        if ret.returncode:
            raise ValueError(ret.stderr)
        else:
            return ret.stdout

    def add(self, *, dry_run: bool, message: str, cat: Category, ref: str, create_branch: bool, add_to_index: bool) -> tuple[Path, str]:
        first_line = message.split('\n')[0] if '\n' in message else message
        slug = self.get_slug((f'{ref}-' if ref else '') + first_line) + '.md'
        branch_name = self.get_branch_name(cat, ref, first_line)
        lines = message.split('\n')[1:] if '\n' in message else []
        if not first_line.startswith('- '):
            first_line = '- ' + first_line
        if ref:
            first_line += f" ({ref})"
        lines.insert(0, first_line)
        md_message = '  \n'.join(line.rstrip() for line in lines)
        log_file = self.cfg.log_dir / cat.name / slug
        log_file.parent.mkdir(parents=True, exist_ok=True)
        if not dry_run:
            if create_branch:
                self.run_command('git', 'switch', '-c', branch_name)
            with log_file.open('w', encoding='utf-8') as f:
                f.write(md_message + '\n')
            if add_to_index:
                self.run_command('git', 'add', str(log_file.relative_to(self.cfg.root_dir)))
        return log_file.relative_to(self.cfg.log_dir), md_message

    def commit(self, *, dry_run: bool, partlog_path: str | None) -> str:
        partlogs = [
            Path(pl[3:])
            for pl in self.run_command(
                'git',
                'status',
                '--porcelain',
                '--untracked-files=no',
                '--no-renames',
                str(self.cfg.log_dir.relative_to(self.cfg.root_dir)),
            ).split('\n')
            if pl.startswith('A')
        ]
        if partlog_path:
            # little dance to ensure to have a relative path from root dir
            partlog = Path(partlog_path).absolute().relative_to(self.cfg.root_dir)
            if partlog not in partlogs:
                raise ValueError(f"{partlog_path} is not amongst added part log files.")
            partlogs.remove(partlog)
        elif len(partlogs) == 0:
            raise ValueError("No part log file found. Cannot commit for you, use `add` first.")
        elif len(partlogs) == 1:
            partlog = partlogs.pop()
        else:
            raise ValueError("Multiple part log files found. Please specify which one is primary as command line paramater.")
        content = (self.cfg.root_dir / partlog).read_text(encoding='utf-8').split('\n')
        m = match(r'^(- )?(.+?)( \((.+)\))?$', content[0])
        _, message, _, ref = m.groups() if m else (None, content[0].lstrip('- '), None, None)
        if ref and ref.startswith('#'):
            ref = f'Ref {ref}'
        commit_message = f'{ref}: {message}\n' if ref else f'{message}\n'
        if content[1:] or partlogs:
            commit_message += '\n'  # blank line
        commit_message += '\n'.join(content[1:])
        for pl in sorted(partlogs):
            commit_message += (self.cfg.root_dir / pl).read_text(encoding='utf-8')
        commit_message = commit_message.strip() + "\n"  # ensure last newline
        if not dry_run:
            self.run_command('git', 'commit', '--file=-', input=commit_message)
        return commit_message

    def list(self) -> list[Path]:
        return [p.relative_to(self.cfg.log_dir) for cat in self.cfg.category_class for p in sorted((self.cfg.log_dir / cat.name).glob('*.md'))]

    def merge(self, *, dry_run: bool, next_version: str, link: str, subtitle: str, with_date: bool, update_index: bool) -> str:
        title = "## "
        title += f"[{next_version}]" if link else next_version
        if with_date:
            title += f" – {datetime.now(timezone.utc).isoformat().split('T')[0]}"
        lines_to_insert = [title]
        if link:
            lines_to_insert.append(f"[{next_version}]: {link}")
        if subtitle:
            lines_to_insert.append("")
            lines_to_insert.append(subtitle)
        for cat in self.cfg.category_class:
            new_category = True
            for p in sorted((self.cfg.log_dir / cat.name).glob('*.md')):
                if new_category:
                    lines_to_insert.append("")
                    cat_prefix = f"{cat.icon} " if cat.icon else ""
                    lines_to_insert.append(f"### {cat_prefix}{cat.name}")
                    new_category = False
                lines_to_insert.extend(p.read_text(encoding='utf-8').strip().split('\n'))
                if not dry_run:
                    p.unlink()
                    if update_index:
                        self.run_command('git', 'rm', '-f', '--ignore-unmatch', str(p.relative_to(self.cfg.root_dir)))
        if self.cfg.log_file.exists():
            lines = self.cfg.log_file.read_text(encoding='utf-8').strip().split('\n')
        else:
            if not dry_run:
                self.cfg.log_file.write_text(self.cfg.default_changelog, encoding='utf-8')
            lines = self.cfg.default_changelog.strip().split('\n')
        if len(lines_to_insert) >= 2:
            pos = 0
            for i, line in enumerate(lines):
                if line.startswith('## '):
                    pos = i
                    break
            else:
                pos = -1
            if pos == -1:
                lines.append("")
                lines.extend(lines_to_insert)
            else:
                lines_to_insert.append("")
                lines = lines[0:pos] + lines_to_insert + lines[pos:]
        changelog = '\n'.join(lines) + '\n'
        if not dry_run:
            self.cfg.log_file.write_text(changelog, encoding='utf-8')
            if update_index:
                self.run_command('git', 'add', str(self.cfg.log_file.relative_to(self.cfg.root_dir)))
        return changelog
